#include"sim.h"
#include"flow.h"

static int num_flows = 0;
static flow_t* flows[MAX_FLOWS];

boolean flow_eq(void* val1, void* val2)
{
    flow_t* flow1 = (flow_t*) val1,
          * flow2 = (flow_t*) val2;

    assert(flow1 != NULL && flow2 != NULL);
    if(flow1 == flow2) return true;

    if(flow1->protocol == flow2->protocol &&
       flow1->src      == flow2->src      &&
       flow1->dst      == flow2->dst      &&
       flow1->src_port == flow2->src_port &&
       flow1->dst_port == flow2->dst_port)
        return true;

    return false;
}

void flow_add(flow_t* flow)
{
    assert(num_flows < MAX_FLOWS);
    flows[num_flows++] = flow;
}

void flow_remove(flow_t* flow)
{
    int i;
    for(i=0; i < num_flows; i++)
    {
        if(flow_eq(flow, flows[i]))
        {
            flow_t* tmp = flows[i];
            flows[i] = flows[num_flows-1];
            num_flows--;
            free(tmp);
            return;
        }
    }

    fprintf(stderr, "ERROR: ");
    flow_print(stderr, flow);
    fprintf(stderr, " is not being inspected.\n");
}

flow_t* flow_zero()
{
    flow_t* ret = (flow_t*)malloc(sizeof(flow_t));

    ret->protocol = 0;
    ret->src      = 0;
    ret->dst      = 0;
    ret->src_port = 0;
    ret->dst_port = 0;

    return ret;
}

flow_t* flow_from_file(FILE* file)
{
    char    protocol[256];
    flow_t* flow = (flow_t*)malloc(sizeof(flow_t));
    u_int   src1, src2, src3, src4,
            dst1, dst2, dst3, dst4;

    fscanf(file, "(%d) %d.%d.%d.%d (%d) > %d.%d.%d.%d (%d)",
                 &(flow->protocol),
                 &src1, &src2, &src3, &src4,
                 &(flow->src_port),
                 &dst1, &dst2, &dst3, &dst4,
                 &(flow->dst_port));

    flow->src = (src1 << 24) |
                (src2 << 16) |
                (src3 << 8)  |
                 src4;

    flow->dst = (dst1 << 24) |
                (dst2 << 16) |
                (dst3 << 8)  |
                 dst4;

    return flow;
}

// Returns a flow if length protocol is UDP | TCP
flow_t* flow_from_packet(const u_char* packet, unsigned int len)
{
    int ipl = packet[ip_offset+0] & 0xF;
    if(ipl == 0)
        return NULL;
    else if(packet[ip_offset+9] == 6  ||     // TCP
            packet[ip_offset+9] == 17)       // UDP
    {
        flow_t* flow    = (flow_t*)malloc(sizeof(flow_t));
        flow->protocol  = packet[ip_offset+9];


        flow->src = (packet[ip_offset+12] << 24) |
                    (packet[ip_offset+13] << 16) |
                    (packet[ip_offset+14] << 8)  |
                    (packet[ip_offset+15]);

        flow->dst = (packet[ip_offset+16] << 24) |
                    (packet[ip_offset+17] << 16) |
                    (packet[ip_offset+18] << 8)  |
                    (packet[ip_offset+19]);

        flow->src_port = (packet[ip_offset+4*ipl] << 8) |
                         (packet[ip_offset+4*ipl+1]);
        flow->dst_port = (packet[ip_offset+4*ipl+2] << 8) |
                         (packet[ip_offset+4*ipl+3]);

        return flow;
    }
    else
        return NULL;
}

flow_t* flow_from_string(char* string)
{
    char    protocol[256];
    flow_t* flow = (flow_t*)malloc(sizeof(flow_t));
    u_int   src1, src2, src3, src4,
            dst1, dst2, dst3, dst4;
   
    sscanf(string, "(%d) %d.%d.%d.%d (%d) > %d.%d.%d.%d (%d)",
                   &(flow->protocol),
                   &src1, &src2, &src3, &src4,
                   &(flow->src_port),
                   &dst1, &dst2, &dst3, &dst4,
                   &(flow->dst_port));

    flow->src = (src1 << 24) |
                (src2 << 16) |
                (src3 << 8)  |
                 src4;

    flow->dst = (dst1 << 24) |
                (dst2 << 16) |
                (dst3 << 8)  |
                 dst4;

    return flow;
}

boolean flow_match(flow_t* flow, const u_char* packet, unsigned int len)
{
    flow_t* pflow = flow_from_packet(packet, len);
    boolean ret = false;
    if(pflow == NULL) return false;

    ret = flow_eq(flow, pflow);
    free(pflow);
    return ret;
}

/* These no longer work due to the change in packet cache structure.
void dict_flow_packet_add(hashtable_t* table, qcap_packet_t* packet)
{
    flow_t* flow = flow_from_packet(packet);
    int* val = NULL;
    if(flow == NULL) flow = flow_zero();

    if((val = hashtable_search(table, flow)) == NULL)   
    {
        val = (int*)malloc(sizeof(int));
        (*val) = 1;
        if(!hashtable_insert(table, flow, val))
        {
            fprintf(stderr, "HASHTABLE INSERT FAILED!\n");
            exit(1);
        }
    }
    else
        (*val)++; 
}

int dict_flow_print(FILE* file, hashtable_t* table)
{
    hashtable_itr_t* itr = hashtable_iterator(table);
    int totals = 0;
    do
    {
        flow_t* flow = (flow_t*) hashtable_iterator_key(itr);
        int*    val  = (int*) hashtable_iterator_value(itr);
        fprintf(file, "%d: ", *val);
        flow_print(file, flow);
        fprintf(file, "\n");
        totals++;
    }
    while(hashtable_iterator_advance(itr));
    free(itr);
    return totals;
} 
*/

boolean flow_matches(const u_char* packet, unsigned int len)
{
    if(num_flows > 0)
    {
        int i;
        flow_t* flow = flow_from_packet(packet, len);

        if(flow == NULL) return false;

        for(i=0; i < num_flows; i++)
        {
            if(flow_eq(flows[i], flow))
            {
                free(flow);
                return true;
            }
        }

        free(flow);
    }

    return false;
}

flow_t* flow_new(int protocol, int src, int dst, int src_port, int dst_port)
{
    flow_t* flow    = (flow_t*)malloc(sizeof(flow_t));
    flow->protocol  = protocol;
    flow->src       = src;
    flow->dst       = dst;
    flow->src_port  = src_port;
    flow->dst_port  = dst_port;
    return flow;
}

void flow_print(FILE* file, flow_t* flow)
{
    u_char src1 = (flow->src & 0xFF000000) >> 24,
           src2 = (flow->src & 0x00FF0000) >> 16,
           src3 = (flow->src & 0x0000FF00) >> 8,
           src4 = (flow->src & 0x000000FF);

    u_char dst1 = (flow->dst & 0xFF000000) >> 24,
           dst2 = (flow->dst & 0x00FF0000) >> 16,
           dst3 = (flow->dst & 0x0000FF00) >> 8,
           dst4 = (flow->dst & 0x000000FF);

    char*   p = "UNKNOWN";
    if(flow->protocol == 1)
        p = "ICMP";
    else if (flow->protocol == 6)
        p = "TCP";
    else if (flow->protocol == 17)
        p = "UDP";
    else if (flow->protocol == 41)
        p = "IPv6";

    fprintf(file, "(%d) %d.%d.%d.%d [%d] > %d.%d.%d.%d [%d]",
                  flow->protocol,
                  src1, src2, src3, src4, flow->src_port,
                  dst1, dst2, dst3, dst4, flow->dst_port);
}

