# Labels are entities in the database which are a name and a color.  These
# labels can be assigned to a tree for nodes and classifications.  These are
# technically stored as the boolen alegbra of p,n-grams along with a final
# classification.
#
# This class gives functions for assigning labels to nodes in a tree, and for
# finding all the labels assigned to a given node.
class Label < ActiveRecord::Base
  belongs_to :tree
  validates_presence_of :name
  validates_presence_of :color
  validates_format_of :name, :with => /^[^<>&]*$/
  validates_format_of :color, :with => /^\#[0-9a-fA-F]{6}$/

  # Gets a list of all labels that exist for a tree
  # @param tree_id The ID of a tree
  def self.find_all(tree_id)
    find :all, :conditions => "tree_id=#{tree_id}"
  end

  # Assign a label to a node in the tree, and a classification within the node
  # @param tree_id The ID of a tree
  # @param label_to The Label object which we are assigning
  # @param cl The Classification object which we are assigning to
  # @param node The Node object (which belongs to the tree specified by
  #             tree_id) which we are assigning to
  def self.assign(tree_id, label_to, cl, node)
    res = self.find_matches(tree_id, node)
    res.each do |r|
      if r.classification_label == cl.label
        LabelAssignment.delete(r.label_assignment_id)
      end
    end

    return if !label_to # there is no label for it anymore

    # add the new assignment
    id = LabelAssignment.create_assignment(label_to.id, cl.id)
    return if !id

    # add all the p,n-grams that lead to this node
    p = node
    while p.parent
      has = p.leftchild?
      p = p.parent
      pngs = NodePngram.find_for_node(p)
      pngs.each do |png|
        newpng = LabelPngram.new
        newpng.label_assignment_id = id
        newpng.pngram_id = png.pngram_id
        newpng.has_pngram = has
        newpng.save
      end
    end
  end

  # Find all labels assigned to a node in a tree
  # @param tree_id The ID of a tree
  # @param node The Node class (which belongs to the tree specified by
  #             tree_id), which we are finding matches for.   If node is not a
  #             terminal node, then no matches will be returned.
  # @return Returns a hash of key, value pairs.  The key is a classification
  #         name like "ether_ipv4_tcp_http".  The value is a read-only Label
  #         object which is the label assigned to the node for the
  #         classification in the key.
  def self.find_for_node(tree_id, node)
    res = find_matches(tree_id, node)

    # build the results by classification name
    ret = {}
    res.each do |r|
      puts "adding #{r.classification_label}"
      ret[r.classification_label] = r
    end
    
    ret
  end

protected
  # tree_id is the tree we are working with
  # node is a Node class, which we are finding matches for. if node is not a
  #   terminal node, then no matches will be returned
  def self.find_matches(tree_id, node)
    return {} if !node.terminal

    where_clause = nil
    num_pngs = 0

    # get all the p,n-grams that lead to this node
    p = node
    while p.parent
      has = p.leftchild?
      p = p.parent
      pngs = NodePngram.find_for_node(p)
      pngs.each do |png|
        i = png.pngram_id
        if !where_clause
          where_clause = 'WHERE'
        else
          where_clause << ' OR '
        end
        where_clause << "(pngram_id=#{i} AND has_pngram=#{has ? 1 : 0})"
        num_pngs += 1
      end
    end

    # ok............... where do I even begin
    # this finds all of the Labels, which have (match or unmatch) the pngrams
    #  for the given Node.  This means all the labels for a node, for each
    #  classfication found within the node

    # the inner select returns a column of label_assignment_ids that
    #   are label assignments which match the p,n-grams for the node
    # the outer select finds all the labels for the tree that match those
    #   assignments, and returns also the 'clasification label' for each one
    find_by_sql %{
      SELECT labels.id AS id,
             labels.tree_id AS tree_id,
             labels.name AS name,
             labels.color AS color,
             cl.label AS classification_label,
             assign.id AS label_assignment_id
        FROM labels
      INNER JOIN label_assignments AS assign ON
        labels.id = assign.label_id
      INNER JOIN classifications AS cl ON
        assign.classification_id = cl.id
      INNER JOIN
        (SELECT a.id label_assignment_id, COUNT(p.id) as count
           FROM label_assignments AS a
         LEFT JOIN label_pngrams AS p ON
           p.label_assignment_id=a.id
         #{where_clause}
         GROUP BY p.label_assignment_id
         HAVING count=#{num_pngs}
         ORDER BY NULL) AS matches ON
        matches.label_assignment_id = assign.id
      WHERE labels.tree_id=#{tree_id}
    }
  end
end
