# The Traffic class models traffic in a tree over time
class Traffic < ActiveRecord::Base
  set_table_name "traffic"

  summary_cache = {}

  # Get a summary of traffic for a tree over a given time period
  # @param tree The Tree object for the tree
  # @param from_time A Time object specifying the lower bound of the time
  #                  period (inclusive)
  # @param to_time A Time object specifying the upper bound of the time
  #                period (inclusive)
  # @return Returns a hash of key,value pairs.  The key is the node's unique
  #         id in the tree.  The value is a TrafficSummary object which
  #         contains the traffic statistics for the node within the time
  #         range requested.
  def self.summary_for_tree(tree, from_time, to_time)
    ret = {}

    from_str = "%04d-%02d-%02d %02d:%02d:%02d" %
      [from_time.year, from_time.month, from_time.day,
       from_time.hour, from_time.min, from_time.sec]
    to_str = "%04d-%02d-%02d %02d:%02d:%02d" %
      [to_time.year, to_time.month, to_time.day,
       to_time.hour, to_time.min, to_time.sec]

    # get the revision id's that this traffic summary is going to span
    revs = Revision.find(:all,
                         :order => "revision ASC",
                         :conditions => "revision >= '#{from_str}' AND " +
                                        "revision <= '#{to_str}' AND " +
                                        "tree_id=#{tree.id}")
    return ret if revs.empty? # no revisions found in the given range

    # get the max position possible for each revision
    max_pos = TerminalNode.maximum('position',
                                   :joins => 'INNER JOIN revisions AS r ON ' +
                                             'revision_id = r.id',
                                   :conditions =>
                                   "r.revision >= '#{from_str}' AND " +
                                   "r.revision <= '#{to_str}'",
                                   :group => 'revision_id')

    # create temp table
    connection.execute "DROP TEMPORARY TABLE IF EXISTS trafsummary"
    connection.execute %{
      CREATE TEMPORARY TABLE trafsummary
        (node_id INT UNSIGNED,
         terminal_node_id int(10) unsigned)}

    tree.all_node_ids.each do |node_id|
      puts "loading #{from_str} to #{to_str} for node #{node_id}"

      # get the position of this node_id for each revision
      my_i_rev = find_by_sql %{
        SELECT revision_id,position FROM internal_nodes
        INNER JOIN revisions AS r ON revision_id = r.id
        WHERE node_id=#{node_id} AND
          r.revision >= '#{from_str}' AND r.revision <= '#{to_str}'}
      my_t_rev = find_by_sql %{
        SELECT revision_id,position FROM terminal_nodes
          INNER JOIN revisions AS r ON revision_id = r.id
          WHERE node_id=#{node_id} AND
            r.revision >= '#{from_str}' AND r.revision <= '#{to_str}'}
      all_my_rev = my_i_rev.concat my_t_rev

      bigcond = nil

      # for each revision for this node...
      all_my_rev.each do |p|
        # get the position and revision id
        pos = p['position'].to_i
        rev = p['revision_id'].to_i
        # get the maximum possible position of this revison
        max = max_pos[rev].to_i
        # get all the possible position numbers which are at "pos" or children
        # of it
        allpos = all_children_positions(pos, max)

        cond = nil
        # build a where clause for being any of these positions
        allpos.each do |i|
          if !cond then cond = "position=#{i}"
          else cond += " OR position=#{i}"
          end
        end

        # add these positions for this revision to the big where clause
        if cond
          bigcond = if !bigcond then 'WHERE ' else bigcond + ' OR ' end
          bigcond += "(revision_id=#{rev} AND (#{cond}))"
        end
      end

      # get all the actual terminal node id's for these positions in this
      # revision
      if bigcond
        connection.execute %{
            INSERT INTO trafsummary
              SELECT #{node_id}, id FROM terminal_nodes
              #{bigcond}}
      end
    end

    # find which columns we are looking for
    cols = nil
    column_names.each do |name|
      if name[0...5] == "ether"
        if !cols then cols = "SUM(#{name}) AS #{name}"
        else          cols += ", SUM(#{name}) AS #{name}"
        end
      end
    end
    raise "No traffic columns exist?" if !cols

    res = find_by_sql %{
      SELECT s.node_id AS node_id, #{cols} FROM traffic
        INNER JOIN trafsummary AS s ON id=s.terminal_node_id
        GROUP BY s.node_id
        ORDER BY NULL}
    res.each do |r|
      sum = TrafficSummary.new
      r.attributes.each do |name, amt|
        sum.add_traffic_type(name, amt) if name != 'node_id' && amt > 0
      end
      ret[r.node_id.to_i] = sum
    end
    ret
  end

protected
  def self.all_children_positions(pos, max)
    allpos = [pos]
    thispos = [pos]
    nextpos = []
    while !thispos.empty?
      thispos.each do |i|
        nextpos << i*2 if i*2 <= max
        nextpos << i*2+1 if i*2+1 <= max
      end
      allpos = allpos.concat(nextpos)
      thispos = nextpos
      nextpos = []
    end
    allpos
  end
end
