# TrafficSummary is a non-database-model.  It represents traffic statistics for
# a node over a time period.
class TrafficSummary
  attr :sum

  # Create a new TrafficSummary with no traffic within it
  def initialize()
    @sum = 0
    @max = 0
    @details = {}
    @percent_leftover = 0
    @percent_leftover_target = nil
  end

  # Add an amount of traffic to the summary
  # @param name The classification name of the traffic
  # @param amount The volume of traffic to add
  def add_traffic_type(name, amount)
    return if amount <= 0
    @details[name] ||= {}           # initialize if it's new
    @details[name][:amount] ||= 0   # initialize if it's new
    @details[name][:amount] += amount
    @sum += amount
    calc_percents

    # give leftovers to the largest slice
    if !@percent_leftover_target || @details[name][:amount] > @max
      @max = @details[name][:amount]
      @percent_leftover_target = name
    end
  end

  # All of the classification names that this traffic summary contains
  def names
    @details.keys
  end

  # The total traffic for all classifications in this summary
  def total
    @sum
  end

  # The volume of traffic for a given classification
  # @param name The classification name for which to get the traffic volume
  def amount(name)
    @details[name][:amount]
  end

  # The percent of traffic that a classification was for total traffic in the
  # summary
  # @param name The classification name for which to find the percent of
  #             traffic volume
  def percent(name)
    percent = @details[name][:percent]
    percent += @percent_leftover if name == @percent_leftover_target
    percent
  end

  # Add another summary to this summary, and return a new TrafficSummary, which
  # is the sum of the two
  def +(summary)
    t = TrafficSummary.new

    self.names.each do |name|
      t.add_traffic_type(name, self.amount(name))
    end
    summary.names.each do |name|
      t.add_traffic_type(name, summary.amount(name))
    end
    t
  end

  # True if no traffic exists for the summary
  def empty?
    @details.empty?
  end

protected
  # Calculate the percent of total traffic for each classification present in
  # the summary
  def calc_percents
    percent_sum = 0
    @details.each do |name, d|
      d[:percent] = (d[:amount].to_f * 100 / @sum).to_i
      percent_sum += d[:percent]
    end
    @percent_leftover = 100 - percent_sum
  end
end
